enum EDownloadStatus
{
    Downloading,
    Done
}

EDownloadStatus DownloadStatus = EDownloadStatus::Done;

string CurrentlyDownloading = "";

[Setting hidden]
string TitlepackName = "TMLagoon@nadeo";
string tpname = "TMLagoon@nadeo";

[Setting hidden]
string TitlepackUrl = "https://files-v4.live.maniaplanet.com/titles/98a3030fa2/TMLagoon@nadeo.Title.Pack.gbx";
string tpurl = "https://files-v4.live.maniaplanet.com/titles/98a3030fa2/TMLagoon@nadeo.Title.Pack.gbx";

[Setting hidden]
bool IsOpen = true;

void StartDownloadByUrl(const string&in DownloadUrl)
{
    CurrentlyDownloading = DownloadUrl;
    DownloadStatus = EDownloadStatus::Downloading;
    UI::ShowNotification("Downloading from: " + DownloadUrl, "This might take a while", vec4(0, 0, 0, 1));
    auto RequestDownloadTitlepack = Net::HttpRequest();
    RequestDownloadTitlepack.Method = Net::HttpMethod::Get;
    RequestDownloadTitlepack.Url = DownloadUrl;
    RequestDownloadTitlepack.Start();
    while (!RequestDownloadTitlepack.Finished())
    {
        yield();
    }
    RequestDownloadTitlepack.SaveToFile(IO::FromUserGameFolder("Packs\\" + TitlepackName + ".Title.Pack.gbx"));

    DownloadStatus = EDownloadStatus::Done;
    UI::ShowNotification("Done!", TitlepackUrl, vec4(0, 1, 0, 1));
    ReloadPacks();
    CurrentlyDownloading = "";
}

void StartDownload(const string&in TitlepackName)
{
    CurrentlyDownloading = TitlepackName;
    UI::ShowNotification("Downloading: " + TitlepackName, "This might take a while", vec4(0, 0, 0, 1));
    auto RequestDownloadUrl = Net::HttpRequest();
    RequestDownloadUrl.Method = Net::HttpMethod::Get;
    RequestDownloadUrl.Url = "https://maniaplanet.com/webservices/titles/" + TitlepackName;
    RequestDownloadUrl.Start();
    DownloadStatus = EDownloadStatus::Downloading;
    while (!RequestDownloadUrl.Finished())
    {
        yield();
    }
    if (RequestDownloadUrl.ResponseCode() != 200) 
    {
        UI::ShowNotification("Couldn't download: " + TitlepackName, RequestDownloadUrl.ResponseCode() + "", vec4(1, 0, 0, 1));
        DownloadStatus = EDownloadStatus::Done;
        return;
    }
    auto JsonDownloadUrl = RequestDownloadUrl.Json();
    string DownloadUrl = JsonDownloadUrl.Get("download_url");
    
    auto RequestDownloadTitlepack = Net::HttpRequest();
    RequestDownloadTitlepack.Method = Net::HttpMethod::Get;
    RequestDownloadTitlepack.Url = DownloadUrl;
    RequestDownloadTitlepack.Start();
    while (!RequestDownloadTitlepack.Finished())
    {
        yield();
    }
    RequestDownloadTitlepack.SaveToFile(IO::FromUserGameFolder("Packs\\" + TitlepackName + ".Title.Pack.gbx"));
    DownloadStatus = EDownloadStatus::Done;
    UI::ShowNotification("Done!", TitlepackName, vec4(0, 1, 0, 1));
    ReloadPacks();
    CurrentlyDownloading = "";
}

void ReloadPacks()
{
    Fids::UpdateTree(Fids::GetUserFolder(""));
}

void SaveTitlepackFromRequest(ref@ Request)
{
    cast<Net::HttpRequest>(Request).SaveToFile(IO::FromUserGameFolder("Packs\\" + TitlepackName + ".Title.Pack.gbx"));
}

array<string> GetManiaTitlesAsIds()
{
    array<string> Output = {};
    auto PacksFolder = Fids::GetUserFolder("Packs");

    for (uint i = 0; i < PacksFolder.Leaves.Length; i++)
    {
        auto Leaf = PacksFolder.Leaves[i];
        auto Nod = Fids::Preload(Leaf);
        if (Nod is null) continue;

        auto Pack = cast<CPlugFilePack>(Nod);
        if (Pack is null) continue;

        Output.InsertLast(Pack.TitleId);
    }
    return Output;
}

array<string> GetStationsAsIds(MwFastBuffer<CGameStation@> Stations)
{
    array<string> Output = {};
    for (uint i = 0; i < Stations.Length; i++)
    {
        auto Station = Stations[i];
        if (Station.Title !is null) 
        {
            Output.InsertLast(Station.Title.TitleId);
        }
    }
    return Output;
}

array<string> GetAllUndownloadedTitlepacks()
{
    // O(3n)
    array<string> Output = {};

    auto Titles = GetManiaTitlesAsIds();
    auto Stations = GetStationsAsIds(cast<CGameManiaPlanet>(GetApp()).Stations);

    for (uint i = 0; i < Stations.Length; i++)
    {
        auto Station = Stations[i];
        if (Titles.Find(Station) < 0)
        {
            Output.InsertLast(Station);
        }
    }

    return Output;
}

void DownloadAllUndownloadedTitlepacks()
{
    array<string> UndownloadedTitlepacks = GetAllUndownloadedTitlepacks();

    for (uint i = 0; i < UndownloadedTitlepacks.Length; i++)
    {
        auto TitlepackName = UndownloadedTitlepacks[i];
        StartDownload(TitlepackName);
    }
}

void Render()
{
    if (!IsOpen) return;
    if (UI::Begin("\\$fd4" + Icons::Download + "\\$z TitlePackDownloader", IsOpen))
    {
        UI::BeginTabBar("##bar");

            if (UI::BeginTabItem("Via ID"))
            {
                tpname = UI::InputText("Titlepack ID (titlepack@author)", TitlepackName);
                TitlepackName = tpname;
                if (UI::Button("Download " + TitlepackName + ".Title.Pack.gbx") and DownloadStatus != EDownloadStatus::Downloading)
                {
                    startnew(StartDownload, TitlepackName);
                }

            UI::EndTabItem();
            }

            if (UI::BeginTabItem("Via URL"))
            {
                tpurl = UI::InputText("Titlepack download URL", TitlepackUrl);
                TitlepackUrl = tpurl;
                if (UI::Button("Download from " + TitlepackUrl) and DownloadStatus != EDownloadStatus::Downloading)
                {
                    startnew(StartDownloadByUrl, TitlepackUrl);
                }

            UI::EndTabItem();
            }

            if (UI::BeginTabItem("Download all undownloaded titlepacks"))
            {
                if (UI::Button("Begin download") and DownloadStatus != EDownloadStatus::Downloading)
                {
                    startnew(DownloadAllUndownloadedTitlepacks);
                }

            UI::EndTabItem();
            }

        UI::EndTabBar();

        UI::Separator();

        if (DownloadStatus == EDownloadStatus::Downloading)
        {
            UI::Text("Currently downloading: " + CurrentlyDownloading);
        }

        UI::Text(DownloadStatus == EDownloadStatus::Downloading ? "Please wait..." : "Ready!");

        UI::Separator();

        if (UI::Button("Reload titlepacks manually"))
        {
            ReloadPacks();
        }

        UI::End();
    }
}

void RenderMenu()
{
    if (UI::MenuItem((DownloadStatus == EDownloadStatus::Downloading ? "\\$fd4" : "\\$0f0") + Icons::Download + "\\$z TitlePackDownloader", tostring(DownloadStatus), IsOpen))
    {
        IsOpen = !IsOpen;
    }
}